from pyramid.httpexceptions import HTTPNotFound
from requests_oauthlib import OAuth2Session

from .constants import S


class AuthProvider(object):

    def __init__(
            self, request, auth_state, auth_scopes, consumer_key,
            consumer_secret, form_url, token_url, resource_url,
            compliance_fixes):
        redirect_uri = request.route_url(
            'authorizations_enter_callback', provider_name=self.name)
        auth_session = OAuth2Session(
            consumer_key,
            scope=auth_scopes,
            redirect_uri=redirect_uri,
            state=auth_state)
        for f in compliance_fixes:
            auth_session = f(auth_session)
        self._consumer_secret = consumer_secret
        self._form_url = form_url
        self._token_url = token_url
        self._resource_url = resource_url
        self._auth_session = auth_session
        self._request_url = request.url

    @property
    def name(self):
        return self.__class__.__name__.lower()

    @property
    def form_url(self):
        return self._auth_session.authorization_url(self._form_url)[0]

    def get_user_definition(self):
        d = self._auth_session.fetch_token(
            self._token_url,
            client_secret=self._consumer_secret,
            authorization_response=self._request_url,
            verify=S['verify_tls'])
        access_token = d['access_token']
        refresh_token = d.get('refresh_token')
        response = self._auth_session.get(self._resource_url)
        resource_definition = response.json()
        return self._get_user_definition(
            resource_definition, access_token, refresh_token)


class Google(AuthProvider):

    def _get_user_definition(
            self, resource_definition, access_token, refresh_token):
        return {
            'name': resource_definition['name'],
            'email': resource_definition['email'],
            'imageUrl': resource_definition['picture'],
        }


'''
# TODO: Fix
# https://docs.microsoft.com/en-us/linkedin/shared/references/v2/profile/profile-picture

class LinkedIn(AuthProvider):

    def _get_user_definition(
            self, resource_definition, access_token, refresh_token):
        first_name = resource_definition['localizedFirstName']
        last_name = resource_definition['localizedLastName']
        return {
            'name': f'{first_name} {last_name}',
            'email': resource_definition['emailAddress'],
            'imageUrl': resource_definition['profilePicture']['displayImage'],
        }
'''


def get_auth_provider(request, auth_state):
    matchdict = request.matchdict
    provider_name = matchdict['provider_name']
    try:
        provider_definition = S['provider_definitions'][provider_name]
    except KeyError:
        raise HTTPNotFound({'provider_name': 'is invalid'})
    return PROVIDER_BY_NAME[provider_name](
        request, auth_state, **provider_definition)


def get_enter_url_by_name(request):
    return {provider_name: request.route_url(
        'authorizations_enter',
        provider_name=provider_name,
    ) for provider_name in S['provider_definitions']}


PROVIDER_BY_NAME = {
    'google': Google,
    # 'linkedin': LinkedIn,
}



import functools
from datetime import datetime
from inspect import getcallargs
from invisibleroads_macros_security import make_random_string
from invisibleroads_posts.views import expect_integer, expect_value
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPInternalServerError,
    HTTPSeeOther,
    HTTPUnauthorized)
from pyramid.security import forget, remember, NO_PERMISSION_REQUIRED
from pyramid.view import view_config

from . import models as M
from .constants import L, S, USER_DEFINITION
from .events import UserAdded
from .providers import (
    get_auth_provider,
    get_enter_url_by_name)


@view_config(
    route_name='authorizations.json',
    renderer='json',
    request_method='GET',
    permission=NO_PERMISSION_REQUIRED)
def see_authorizations_json(request):
    session = request.session
    d = {
        'urls': {
            'enter': get_enter_url_by_name(request),
            'leave': request.route_url('authorizations_leave'),
        },
    }
    if 'user' in session:
        user_definition = session['user']
        d['user'] = {_: user_definition.get(_) for _ in S['public_attributes']}
    return d


@view_config(
    route_name='authorizations_enter',
    permission=NO_PERMISSION_REQUIRED)
def see_provider(request):
    target_url = get_target_url(request)

    if S['mock']:
        params = request.params
        user_email = params.get('email', USER_DEFINITION['email'])
        user_definition = dict(USER_DEFINITION, **{
            'name': user_email.split('@')[0].title(),
            'email': user_email,
            'imageUrl': S['image_url'],
        })
        return welcome_user(request, user_definition, target_url)

    form_url = get_form_url(request, target_url)
    return HTTPSeeOther(form_url)


@view_config(
    route_name='authorizations_enter_callback',
    permission=NO_PERMISSION_REQUIRED)
def remember_user(request):
    params = request.params
    session = request.session
    auth_state = session.pop('auth_state', '')
    target_url = session.pop('target_url', '/')

    if not params.get('code') or params.get('state') != auth_state:
        raise HTTPSeeOther(target_url)

    auth_provider = get_auth_provider(request, auth_state)
    user_definition = auth_provider.get_user_definition()
    return welcome_user(request, user_definition, target_url)


@view_config(
    route_name='authorizations_leave',
    permission=NO_PERMISSION_REQUIRED)
def forget_user(request):
    target_url = get_target_url(request)
    session = request.session
    session.invalidate()
    headers = forget(request)
    return HTTPSeeOther(target_url, headers=headers)


def get_form_url(request, target_url):
    auth_state = make_random_string(S['auth_state_length'])
    auth_provider = get_auth_provider(request, auth_state)
    session = request.session
    session['auth_state'] = auth_state
    session['target_url'] = target_url
    return auth_provider.form_url


def get_target_url(request):
    params = request.params
    return params.get('targetUrl', S['target_url']).strip()


def welcome_user(request, user_definition, target_url):
    user_name = user_definition['name']
    user_email = user_definition['email']

    if S['storage'] == 'database':
        User = M.User
        is_new_user = False
        db = request.db
        user = db.query(User).filter_by(email=user_email).first()
        if not user:
            user = User.make_unique_record(db)
            user.email = user_email
            is_new_user = True
        user.name = user_name
        user.image_url = user_definition.get('imageUrl', S['image_url'])
        user.modification_datetime = datetime.utcnow()
        if is_new_user:
            request.registry.notify(UserAdded(user, request))
        user_id = user.id
    else:
        user_id = user_definition.get('id', user_email)

    # Pass user_definition to UserAuthService while remembering user_id
    user_definition['id'] = user_id
    request.user_definition = user_definition
    headers = remember(request, user_id)

    # Store user_definition in session
    session = request.session
    session['user'] = user_definition
    return HTTPSeeOther(target_url, headers=headers)


def handle_redis_connection_error(request):
    L.error('redis: is not accessible')
    raise HTTPInternalServerError({})


def handle_csrf_origin_error(request):
    raise HTTPBadRequest({'csrf': 'has bad origin'})


def handle_csrf_token_error(request):
    raise HTTPBadRequest({'csrf': 'has bad token'})


def authorize_param(f):
    @functools.wraps(f)
    def wrapper(*args, **kw):
        value = f(*args, **kw)
        args = getcallargs(f, *args, **kw)
        request = args['request']
        default = args['default']
        if not request.authenticated_userid and value != default:
            raise HTTPUnauthorized
        return value
    return wrapper


authorize_value = authorize_param(expect_value)
authorize_integer = authorize_param(expect_integer)



from invisibleroads_macros_configuration import (
    load_attribute,
    load_attributes,
    set_default)
from invisibleroads_macros_security import make_random_string
from invisibleroads_records.models import Base
from os import environ
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.exceptions import BadCSRFOrigin, BadCSRFToken
from pyramid.security import Allow, Authenticated
from pyramid.settings import asbool, aslist
from pyramid_authsanity.interfaces import IAuthService
from pyramid_redis_sessions import session_factory_from_settings
from redis import ConnectionError as RedisConnectionError

from . import models as M
from .constants import (
    DEFAULT_SECRET,
    DEFAULT_SECRET_ERROR_MESSAGE,
    INVISIBLEROADS_USERS_SETTINGS_PREFIX,
    L,
    REDIS_SESSIONS_COOKIE_NAME,
    REDIS_SESSIONS_PREFIX,
    REDIS_SESSIONS_SECRET_LENGTH,
    REDIS_SESSIONS_SETTINGS_PREFIX,
    REDIS_SESSIONS_TIMEOUT_IN_SECONDS,
    S)
from .routines import get_crypt
from .services import make_user_auth_service
from .views import (
    handle_csrf_origin_error,
    handle_csrf_token_error,
    handle_redis_connection_error)


def includeme(config):
    configure_settings(config)
    configure_request_session_factory(config)
    configure_security_policy(config)
    configure_provider_definitions(config)
    configure_views(config)


def configure_settings(config, prefix=INVISIBLEROADS_USERS_SETTINGS_PREFIX):
    settings = config.get_settings()
    UserMixin = S.set(
        settings, prefix, 'user_mixin', M.UserMixin, load_attribute)
    if not hasattr(M, 'User'):
        M.User = type('User', (UserMixin, Base), {})
    S.set(settings, prefix, 'mock', S['mock'], asbool)
    S.set(settings, prefix, 'storage', S['storage'])
    S.set(settings, prefix, 'cookie_secure', S['cookie_secure'], asbool)
    S.set(settings, prefix, 'cookie_httponly', S['cookie_httponly'], asbool)
    S.set(settings, prefix, 'verify_tls', S['verify_tls'], asbool)
    S.set(settings, prefix, 'auth_state_length', S['auth_state_length'], int)
    S.set(settings, prefix, 'public_attributes', S[
        'public_attributes'], aslist)
    S.set(settings, prefix, 'secret', S['secret'])
    if S['secret'] == DEFAULT_SECRET:
        L.warning(DEFAULT_SECRET_ERROR_MESSAGE)
    S.set(settings, prefix, 'default_permission', S[
        'default_permission'], asbool)
    S.set(settings, prefix, 'require_csrf', S['require_csrf'], asbool)
    S.set(settings, prefix, 'target_url', S['target_url'])
    S.set(settings, 'redis.users.', 'prefix', S['redis.users.prefix'])


def configure_request_session_factory(
        config, prefix=REDIS_SESSIONS_SETTINGS_PREFIX):
    settings = config.get_settings()
    set_default(settings, prefix + 'secret', make_random_string(
        REDIS_SESSIONS_SECRET_LENGTH))
    set_default(
        settings, prefix + 'timeout', REDIS_SESSIONS_TIMEOUT_IN_SECONDS)
    set_default(settings, prefix + 'cookie_name', REDIS_SESSIONS_COOKIE_NAME)
    set_default(settings, prefix + 'cookie_secure', S['cookie_secure'])
    set_default(settings, prefix + 'cookie_httponly', S['cookie_httponly'])
    set_default(settings, prefix + 'prefix', REDIS_SESSIONS_PREFIX)
    config.set_session_factory(session_factory_from_settings(settings))
    config.add_view(
        handle_redis_connection_error, context=RedisConnectionError)


def configure_security_policy(config):
    config.set_authorization_policy(ACLAuthorizationPolicy())
    config.include('pyramid_authsanity')
    config.register_service_factory(make_user_auth_service, iface=IAuthService)
    config.set_default_permission(S['default_permission'])
    config.set_default_csrf_options(require_csrf=S['require_csrf'])
    config.add_view(handle_csrf_origin_error, context=BadCSRFOrigin)
    config.add_view(handle_csrf_token_error, context=BadCSRFToken)
    S['crypt'] = get_crypt()


def configure_provider_definitions(
        config, prefix=INVISIBLEROADS_USERS_SETTINGS_PREFIX):
    settings = config.get_settings()
    S['provider_definitions'] = d = {}
    for provider_name in set_default(
            settings, prefix + 'auth_providers', [], aslist):
        provider_prefix = prefix + 'auth_providers.' + provider_name + '.'
        d[provider_name] = {
            'auth_scopes': aslist(settings[provider_prefix + 'auth_scopes']),
            'consumer_key': settings[provider_prefix + 'consumer_key'],
            'consumer_secret': settings[provider_prefix + 'consumer_secret'],
            'form_url': settings[provider_prefix + 'form_url'],
            'token_url': settings[provider_prefix + 'token_url'],
            'resource_url': settings[provider_prefix + 'resource_url'],
            'compliance_fixes': load_attributes(settings[
                provider_prefix + 'compliance_fixes']),
        }
    if not S['cookie_secure']:
        environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'


def configure_views(config):

    class DefaultRootFactory(object):

        __acl__ = [
            (Allow, Authenticated, S['default_permission']),
        ]

        def __init__(self, request):
            pass

    config.set_root_factory(DefaultRootFactory)
    config.include('.routes')
    config.scan('.views')



from invisibleroads_records.models import (
    CreationMixin,
    ModificationMixin,
    RecordMixin)
from sqlalchemy import Column
from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy.types import LargeBinary

from .routines import decrypt, encrypt


class CaseInsensitiveString(str):

    def __lt__(self, other):
        return self.casefold() < other.casefold()

    def __le__(self, other):
        return self.casefold() <= other.casefold()

    def __eq__(self, other):
        return self.casefold() == other.casefold()

    def __ne__(self, other):
        return self.casefold() != other.casefold()

    def __gt__(self, other):
        return self.casefold() > other.casefold()

    def __ge__(self, other):
        return self.casefold() >= other.casefold()


class CaseInsensitiveEncryptComparator(Comparator):

    def operate(self, op, other, **kwargs):
        return op(self.__clause_element__(), encrypt(
            other.casefold()), **kwargs)


class EncryptedNameMixin(object):
    'Mixin class for a case-insensitive encrypted name'
    _name = Column(LargeBinary)

    @hybrid_property
    def name(self):
        return CaseInsensitiveString(decrypt(self._name))

    @name.setter
    def name(self, name):
        self._name = encrypt(name.casefold())

    @name.comparator
    def name(Class):
        return CaseInsensitiveEncryptComparator(Class._name)


class EncryptedEmailMixin(object):
    'Mixin class for a case-insensitive encrypted email address'
    _email = Column(LargeBinary)

    @hybrid_property
    def email(self):
        return CaseInsensitiveString(decrypt(self._email))

    @email.setter
    def email(self, email):
        self._email = encrypt(email.casefold())

    @email.comparator
    def email(Class):
        return CaseInsensitiveEncryptComparator(Class._email)


class EncryptedImageUrlMixin(object):
    'Mixin class for a case-insensitive encrypted image URL'
    _image_url = Column(LargeBinary)

    @hybrid_property
    def image_url(self):
        return CaseInsensitiveString(decrypt(self._image_url))

    @image_url.setter
    def image_url(self, image_url):
        self._image_url = encrypt(image_url.casefold())

    @image_url.comparator
    def image_url(Class):
        return CaseInsensitiveEncryptComparator(Class._image_url)


class UserMixin(
        EncryptedNameMixin,
        EncryptedEmailMixin,
        EncryptedImageUrlMixin,
        ModificationMixin,
        CreationMixin,
        RecordMixin):
    __tablename__ = 'user'



from pyramid import testing
from pytest import fixture

from . import models as M


@fixture
def users_request(users_config, database, data_folder):
    request = testing.DummyRequest(
        db=database,
        data_folder=data_folder)
    request.json_body = {}
    yield request


@fixture
def users_config(users_settings):
    config = testing.setUp(settings=users_settings)
    config.include('invisibleroads_posts')
    config.include('invisibleroads_records')
    config.include('invisibleroads_users')
    yield config
    testing.tearDown()


@fixture
def users_settings(records_settings):
    users_settings = records_settings
    yield users_settings


@fixture
def user(users_request):
    database = users_request.database
    User = M.User
    user = User.make_unique_record(database)
    user.name = 'User'
    user.email = 'user@example.com'
    return user
